import operator
import logging
from importlib import import_module

try:
    from rita.engine.translate_spacy import compile_rules as spacy_engine
except ImportError:
    pass

from rita.engine.translate_standalone import compile_rules as standalone_engine

from rita.utils import SingletonMixin


logger = logging.getLogger(__name__)


class Config(SingletonMixin):
    def __init__(self):
        self.available_engines = []
        self.engines_by_key = {}
        self.current_engine = None

        try:
            self.register_engine(1, "spacy", spacy_engine)
        except NameError:
            # spacy_engine is not imported
            pass
        self.register_engine(2, "standalone", standalone_engine)

    def register_engine(self, priority, key, compile_fn):
        self.available_engines.append((priority, key, compile_fn))
        self.engines_by_key[key] = compile_fn
        sorted(self.available_engines, key=operator.itemgetter(0))

    @property
    def default_engine(self):
        (_, key, compile_fn) = self.available_engines[0]
        self.current_engine = key
        return compile_fn

    def set_engine(self, key):
        self.current_engine = key
        return self.engines_by_key[key]

    @property
    def list_branching(self):
        if self.current_engine == "spacy":
            return True

        return False


class SessionConfig(object):
    def __init__(self):
        self._root = Config()
        self.modules = []
        # Default config
        self._data = {
            "ignore_case": True,
            "implicit_punct": True,
            "deaccent": True,
            "implicit_hyphon": False,
        }
        self.variables = {}

    def register_module(self, mod_name):
        logger.debug("Importing module: {}".format(mod_name))
        self.modules.append(import_module(mod_name))

    def set_variable(self, k, v):
        self.variables[k] = v

    def get_variable(self, k):
        return self.variables[k]

    def __getattr__(self, name):
        if name == "_root":
            return self._root

        elif name in self._data:
            return self._data[name]

        return getattr(self._root, name)

    def set_config(self, k, v):
        # Handle booleans first
        if v.upper() in ["1", "T", "Y"]:
            self._data[k] = True
        elif v.upper() in ["0", "F", "N"]:
            self._data[k] = False
        else:
            self._data[k] = v


def with_config(fn):
    def wrapper(*args, **kwargs):
        config = SessionConfig()
        return fn(*args, config=config, **kwargs)

    return wrapper



import logging

from functools import partial

logger = logging.getLogger(__name__)


def any_of_parse(lst, config, op=None):
    if config.ignore_case:
        normalized = sorted([item.lower()
                             for item in lst])
        base = {"LOWER": {"REGEX": r"^({0})$".format("|".join(normalized))}}
    else:
        base = {"TEXT": {"REGEX": r"^({0})$".format("|".join(sorted(lst)))}}

    if op:
        base["OP"] = op
    yield base


def regex_parse(r, config, op=None):
    if config.ignore_case:
        d = {"LOWER": {"REGEX": r.lower()}}
    else:
        d = {"TEXT": {"REGEX": r}}

    if op:
        d["OP"] = op
    yield d


def fuzzy_parse(r, config, op=None):
    # TODO: build premutations
    d = {"LOWER": {"REGEX": "({0})[.,?;!]?".format("|".join(r))}}
    if op:
        d["OP"] = op
    yield d


def generic_parse(tag, value, config, op=None):
    d = {}
    if tag == "ORTH" and config.ignore_case:
        d["LOWER"] = value.lower()
    else:
        d[tag] = value

    if op:
        d["OP"] = op
    yield d


def punct_parse(_, config, op=None):
    d = {}
    d["IS_PUNCT"] = True
    if op:
        d["OP"] = op
    yield d


def phrase_parse(value, config, op=None):
    """
    TODO: Does not support operators
    """
    splitter = next((s for s in ["-", " "]
                     if s in value), None)
    if splitter:
        buff = value.split(splitter)
        yield next(generic_parse("ORTH", buff[0], config=config, op=None))
        for b in buff[1:]:
            if splitter != " ":
                yield next(generic_parse("ORTH", splitter, config=config, op=None))
            yield next(generic_parse("ORTH", b, config=config, op=None))
    else:
        yield generic_parse("ORTH", value, config=config, op=None)


def tag_parse(r, config, op=None):
    """
    For generating POS/TAG patterns based on a Regex
    e.g. TAG("^NN|^JJ") for adjectives or nouns
    """
    d = {"TAG": {"REGEX": r}}

    if op:
        d["OP"] = op
    yield d


PARSERS = {
    "any_of": any_of_parse,
    "value": partial(generic_parse, "ORTH"),
    "regex": regex_parse,
    "entity": partial(generic_parse, "ENT_TYPE"),
    "lemma": partial(generic_parse, "LEMMA"),
    "pos": partial(generic_parse, "POS"),
    "punct": punct_parse,
    "fuzzy": fuzzy_parse,
    "phrase": phrase_parse,
    "tag": tag_parse,
}


def rules_to_patterns(label, data, config):
    logger.debug(data)
    return {
        "label": label,
        "pattern": [p
                    for (t, d, op) in data
                    for p in PARSERS[t](d, config=config, op=op)],
    }


def compile_rules(rules, config):
    logger.info("Using spaCy rules implementation")
    return [rules_to_patterns(*group, config=config)
            for group in rules]



import sys
import logging
try:
    import inflect
except ImportError:
    logging.exception(
        "Pluralize module requires 'inflect' package to be installed."
        "Install it and try again"
    )
    sys.exit(1)

from rita.macros import resolve_value
from rita.utils import flatten


def pluralizing(initial_list):
    """"
    For a list of nouns, it will return a list of the plurals and the initial nouns
    """
    p = inflect.engine()
    plurals = [p.plural(word) for word in initial_list]
    return initial_list + plurals


def PLURALIZE(*args, config, op=None):
    """
    For a noun or a list of nouns, it will match any singular or plural word
    Usage for a single word, e.g.:
    PLURALIZE("car")
    Usage for lists, e.g.:
    vehicles = {"car", "bicycle", "ship"}
    PLURALIZE(vehicles)
    Will work even for regex or if the lemmatizer of spaCy is making an error
    Has dependency to the Python inflect package https://pypi.org/project/inflect/
    """
    if type(args[0]) == list:
        initial_list = [resolve_value(arg, config=config)
                        for arg in flatten(args)]
    else:
        initial_list = [args[0]]
    return "any_of", pluralizing(initial_list), op



import re
import logging

from functools import partial


logger = logging.getLogger(__name__)


MAX_DEPTH = 5
ALIAS_PATTERN = re.compile(r"@alias\s+(?P<original>(\w|[_])+)\s+(?P<alias>(\w|[_])+)")


def handle_import(m, depth=0):
    path = m.group("path")
    logger.debug("Importing: {}".format(path))
    with open(path, "r") as f:
        return precompile(f.read(), depth+1)


def precompile(raw, depth=0):
    if depth > MAX_DEPTH:
        raise RuntimeError(
            "Maximum depth limit has been reached. "
            "Please check if you don't have cyclical imports"
        )

    raw = re.sub(
        r"@import\s+[\"'](?P<path>(\w|[/\-.])+)[\"']",
        partial(handle_import, depth=depth),
        raw
    )

    for m in ALIAS_PATTERN.finditer(raw):
        # Delete alias definition
        full = m.group(0)
        raw = raw.replace(full, "")

        original = m.group("original")
        alias = m.group("alias")
        raw = re.sub(r"(?:(\s|->|{{))(?P<alias>{})([\(])".format(alias), r"\1{}(".format(original), raw)

    return raw



import logging

from functools import reduce

from rita.utils import Node, deaccent
from rita.macros import resolve_value

logger = logging.getLogger(__name__)


def handle_prefix(rules, config):
    """
    If we have a prefix and rule following it, we apply this prefix on that rule
    """
    def apply_prefix(pattern, prefix):
        (name, args, op) = pattern
        if name == "any_of":
            return (name, list(["{0}{1}".format(prefix, item)
                                for item in args]), op)
        elif name == "value":
            return name, "{0}{1}".format(prefix, args), op
        else:
            logger.warning("Don't know how to apply prefix on: {}".format(name))
            return pattern

    def gen():
        prefix = None
        for p in pattern:
            (name, args, op) = p
            if name == "prefix":
                prefix = args
            else:
                if prefix:
                    yield apply_prefix(p, prefix)
                    prefix = None
                else:
                    yield p
    for group_label, pattern in rules:
        yield group_label, list(gen())


def handle_deaccent(rules, config):
    """
    If we get accented word, eg: {WORD("naïve"), WORD("bayes")}
    In case of word, it should become list => {IN_LIST({"naïve", "naive"}), WORD("bayes")}
    In case of list, it should extend list with extra items for accented and not accented versions
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                (name, args, op) = p
                if name == "value":
                    (v1, v2) = (args, deaccent(args))
                    if v1 != v2:
                        yield "any_of", (v1, v2,), op
                    else:
                        yield p
                elif name == "any_of":
                    def items():
                        for w in args:
                            (v1, v2) = (w, deaccent(w))
                            if v1 != v2:
                                yield v1
                                yield v2
                            else:
                                yield v1

                    yield "any_of", list(items()), op
                else:
                    yield p

        yield group_label, list(gen())


def add_implicit_punct(rules, config):
    """
    When writing rule,
    user usually doesn't care about some punct characters between words.
    We add them implicitly (unless this setting is turned off)
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                yield p
                yield "punct", None, "?"

        if len(pattern) == 1:
            yield group_label, pattern
        else:
            yield group_label, list(gen())[:-1]


def add_implicit_hyphon(rules, config):
    """
    When writing rule,
    user usually doesn't care about hyphon characters - between words.
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                yield p
                yield "value", "-", "?"

        if len(pattern) == 1:
            yield group_label, pattern
        else:
            yield group_label, list(gen())[:-1]


def handle_multi_word(rules, config):
    """
    spaCy splits everything in tokens.
    Words with dash ends up in different tokens.

    We don't want for user to even care about this,
    so we make this work implicitly

    WORD("Knee-length") => WORD("Knee"), WORD("-"), WORD("length")
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                (name, args, op) = p
                if name == "value" and is_complex(args):
                    yield "phrase", args, op
                else:
                    yield p

        yield group_label, list(gen())


def is_complex(arg):
    # if we want to use `-` as a word
    if arg.strip() == "-":
        return False

    splitters = ["-", " "]
    return any([s in arg
                for s in splitters])


def has_complex(args):
    """
    Tells if any of arguments will be impacted by tokenizer
    """
    return any([is_complex(a)
                for a in args])


def branch_pattern(pattern, config):
    """
    Creates multiple lists for each possible permutation
    """
    root = Node()
    current = root
    depth = 0
    for idx, p in enumerate(pattern):
        if p[0] == "either":
            n = Node()
            current.add_next(n)
            current = n
            current.depth = depth
            for e in p[1]:
                current.add_child(e(config=config))
                depth += 1
        elif p[0] == "any_of" and has_complex(p[1]):
            _all = set(p[1])
            _complex = set(filter(is_complex, _all))
            simple = _all - _complex
            n = Node()
            current.add_next(n)
            current = n
            current.depth = depth
            current.add_child(("any_of", simple, p[2]))
            for c in sorted(_complex):
                current.add_child(("phrase", c, p[2]))
                depth += 1
        else:
            n = Node(p)
            current.add_next(n)
            current = n
            current.depth = depth

    for p in root.unwrap():
        yield p


def handle_rule_branching(rules, config):
    """
    If we have an OR statement, eg. `WORD(w1)|WORD(w2)`,
    Generic approach is to clone rules and use w1 in one, w2 in other.
    It may be an overkill, but some situations are not covered
    with simple approach
    """
    for group_label, pattern in rules:
        # Covering WORD(w1)|WORD(w2) case
        if any([p == "either"
                for (p, _, _) in pattern]):
            for p in branch_pattern(pattern, config):
                yield group_label, p

        # Covering case when there are complex items in list
        elif config.list_branching and any([p == "any_of" and has_complex(o)
                                            for (p, o, _) in pattern]):
            for p in branch_pattern(pattern, config):
                yield group_label, p
        else:
            yield group_label, pattern


def dummy(rules, config):
    """
    Placeholder which does nothing
    """
    logger.debug("Initial rules: {}".format(rules))
    return rules


def rule_tuple(d):
    return d["label"], d["data"]


def expand_patterns(rules, config):
    """
    We can have situations where inside pattern we have another pattern (via Variable).
    We want to expand this inner pattern and prepend to outer pattern
    """
    for group_label, pattern in rules:
        def gen():
            for p in pattern:
                if callable(p):
                    yield resolve_value(p, config=config)
                else:
                    yield p

        yield group_label, list(gen())


def preprocess_rules(root, config):
    logger.info("Preprocessing rules")

    rules = [rule_tuple(doc())
             for doc in root
             if doc and doc()]

    pipeline = [dummy, expand_patterns, handle_deaccent, handle_rule_branching, handle_multi_word, handle_prefix]

    if config.implicit_hyphon:
        logger.info("Adding implicit Hyphons")
        pipeline.append(add_implicit_hyphon)
    elif config.implicit_punct:
        logger.info("Adding implicit Punctuations")
        pipeline.append(add_implicit_punct)

    return reduce(lambda acc, p: p(acc, config), pipeline, rules)



import logging
import types

from rita.utils import flatten

logger = logging.getLogger(__name__)


def resolve_value(obj, config):
    logger.debug("Resolving value: {0}".format(obj))

    if isinstance(obj, str):
        return obj

    elif isinstance(obj, tuple):
        return obj

    elif isinstance(obj, list):
        return obj

    elif isinstance(obj, types.GeneratorType):
        return "either", list(obj), None

    return obj(config=config)


def ANY(config, op=None):
    return "regex", r".*", op


def PUNCT(config, op=None):
    return "punct", None, op


def MARK(type_, obj, config, op=None):
    return {
        "label": resolve_value(type_, config=config),
        "data": resolve_value(obj, config=config)
    }


def LOAD(*args, config):
    fpath = resolve_value(args[0], config=config)
    with open(fpath, "r") as f:
        return list([line.strip() for line in f.readlines()])


def ASSIGN(k, v, config, op=None):
    logger.debug("Assigning: {0} -> {1}".format(k, v))
    config.set_variable(k, resolve_value(v, config=config))


def IN_LIST(*args, config, op=None):
    return "any_of", [resolve_value(arg, config=config)
                      for arg in flatten(args)], op


def PATTERN(*args, config, op=None):
    context = []
    for arg in args:
        result = resolve_value(arg, config=config)
        if isinstance(result, list):
            context += result
        else:
            context.append(result)

    return context


def WORD(*args, config, op=None):
    if len(args) == 1:
        literal = resolve_value(args[0], config=config)
        return "value", literal, op
    elif len(args) == 0:
        return "regex", "(\\w|['_-])+", op


def NUM(*args, config, op=None):
    if len(args) == 1:
        literal = resolve_value(args[0], config=config)
        return "value", literal, op
    elif len(args) == 0:
        return "regex", "\\d+[.]?\\d*", op


def POS(name, config, op=None):
    return "pos", resolve_value(name, config=config), op


def LEMMA(name, config, op=None):
    return "lemma", resolve_value(name, config=config), op


def ENTITY(name, config, op=None):
    return "entity", resolve_value(name, config=config), op


def PREFIX(name, config, op=None):
    return "prefix", resolve_value(name, config=config), op


def IMPORT(module, config):
    mod_name = resolve_value(module, config=config)
    config.register_module(mod_name)


def CONFIG(setting, value, config):
    logger.debug("Config {0} -> {1}".format(setting, value))
    config.set_config(setting, resolve_value(value, config=config))


def EXEC(obj, config):
    return resolve_value(obj, config=config)



from rita.macros import resolve_value


def TAG(name, config, op=None):
    """
    For generating POS/TAG patterns based on a Regex
    e.g. TAG("^NN|^JJ") for nouns or adjectives
    """
    return "tag", resolve_value(name, config=config), op



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['rita', 'rita.engine', 'rita.modules']

package_data = \
{'': ['*']}

install_requires = \
['ply==3.11']

entry_points = \
{'console_scripts': ['rita = rita.run:main']}

setup_kwargs = {
    'name': 'rita-dsl',
    'version': '0.5.9',
    'description': 'DSL for building language rules',
    'long_description': '# RITA DSL\n\n[![Documentation Status](https://readthedocs.org/projects/rita-dsl/badge/?version=latest)](http://rita-dsl.readthedocs.io/?badge=latest)\n[![codecov](https://codecov.io/gh/zaibacu/rita-dsl/branch/master/graph/badge.svg)](https://codecov.io/gh/zaibacu/rita-dsl)\n[![made-with-python](https://img.shields.io/badge/Made%20with-Python-1f425f.svg)](https://www.python.org/)\n[![Maintenance](https://img.shields.io/badge/Maintained%3F-yes-green.svg)](https://github.com/zaibacu/rita-dsl/graphs/commit-activity)\n[![PyPI version fury.io](https://badge.fury.io/py/rita-dsl.svg)](https://pypi.python.org/pypi/rita-dsl/)\n[![PyPI download month](https://img.shields.io/pypi/dm/rita-dsl.svg)](https://pypi.python.org/pypi/rita-dsl/)\n[![GitHub license](https://img.shields.io/github/license/zaibacu/rita-dsl.svg)](https://github.com/zaibacu/rita-dsl/blob/master/LICENSE)\n\nThis is a language, loosely based on language [Apache UIMA RUTA](https://uima.apache.org/ruta.html), focused on writing manual language rules, which compiles into either [spaCy](https://github.com/explosion/spaCy) compatible patterns, or pure regex. These patterns can be used for doing [manual NER](https://spacy.io/api/entityruler) as well as used in other processes, like retokenizing and pure matching\n\n## An Introduction Video\n[![Intro](https://img.youtube.com/vi/GScerMeWz68/0.jpg)](https://www.youtube.com/watch?v=GScerMeWz68)\n\n## Links\n- [Live Demo](https://rita-demo.herokuapp.com/)\n- [Documentation](http://rita-dsl.readthedocs.io/)\n- [QuickStart](https://rita-dsl.readthedocs.io/en/latest/quickstart/)\n\n## Support\n\n[![reddit](https://img.shields.io/reddit/subreddit-subscribers/ritaDSL?style=social)](https://www.reddit.com/r/ritaDSL/)\n[![Gitter](https://badges.gitter.im/rita-dsl/community.svg)](https://gitter.im/rita-dsl/community?utm_source=badge&utm_medium=badge&utm_campaign=pr-badge)\n\n## Install\n\n`pip install rita-dsl`\n\n## Simple Rules example\n\n```python\nrules = """\ncuts = {"fitted", "wide-cut"}\nlengths = {"short", "long", "calf-length", "knee-length"}\nfabric_types = {"soft", "airy", "crinkled"}\nfabrics = {"velour", "chiffon", "knit", "woven", "stretch"}\n\n{IN_LIST(cuts)?, IN_LIST(lengths), WORD("dress")}->MARK("DRESS_TYPE")\n{IN_LIST(lengths), IN_LIST(cuts), WORD("dress")}->MARK("DRESS_TYPE")\n{IN_LIST(fabric_types)?, IN_LIST(fabrics)}->MARK("DRESS_FABRIC")\n"""\n```\n\n### Loading in spaCy\n```python\nimport spacy\nfrom rita.shortcuts import setup_spacy\n\n\nnlp = spacy.load("en")\nsetup_spacy(nlp, rules_string=rules)\n```\n\nAnd using it:\n```python\n>>> r = nlp("She was wearing a short wide-cut dress")\n>>> [{"label": e.label_, "text": e.text} for e in r.ents]\n[{\'label\': \'DRESS_TYPE\', \'text\': \'short wide-cut dress\'}]\n```\n\n### Loading using Regex (standalone)\n```python\nimport rita\n\npatterns = rita.compile_string(rules, use_engine="standalone")\n```\n\nAnd using it:\n```python\n>>> list(patterns.execute("She was wearing a short wide-cut dress"))\n[{\'end\': 38, \'label\': \'DRESS_TYPE\', \'start\': 18, \'text\': \'short wide-cut dress\'}]\n```\n\n## Special Thank You\n\nSpecial thanks goes to \n\n[![JetBrains](docs/assets/jetbrains.svg?raw=true)](https://www.jetbrains.com/?from=rita-dsl)\n\nfor supporting development of this library\n',
    'author': 'Šarūnas Navickas',
    'author_email': 'zaibacu@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/zaibacu/rita-dsl',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.5,<4.0',
}


setup(**setup_kwargs)



import logging
import re
import json

from functools import partial
from itertools import groupby

logger = logging.getLogger(__name__)


def apply_operator(syntax, op):
    if not op:
        return syntax

    elif op == "!":  # A bit complicated one
        return (r"((?!{})\w+)".format(syntax
                                      .rstrip(")")
                                      .lstrip("(")))
    else:
        return syntax + op


def any_of_parse(lst, op=None):
    clause = r"((^|\s)(({0})\s?))".format("|".join(sorted(lst, key=lambda x: (-len(x), x))))
    return apply_operator(clause, op)


def regex_parse(r, op=None):
    initial = "(" + r + r"\s?" + ")"
    return apply_operator(initial, op)


def not_supported(key, *args, **kwargs):
    raise RuntimeError(
        "Rule '{0}' is not supported in standalone mode"
        .format(key)
    )


def person_parse(op=None):
    return apply_operator(r"([A-Z]\w+\s?)", op)


def entity_parse(value, op=None):
    if value == "PERSON":
        return person_parse(op)
    else:
        return not_supported(value)


def punct_parse(_, op=None):
    return apply_operator(r"([.,!;?:]\s?)", op)


def word_parse(value, op=None):
    initial = r"({}\s?)".format(value)
    return apply_operator(initial, op)


def fuzzy_parse(r, op=None):
    # TODO: build premutations
    return apply_operator(r"({0})[.,?;!]?".format("|".join(r)), op)


def phrase_parse(value, op=None):
    return apply_operator(r"({}\s?)".format(value), op)


PARSERS = {
    "any_of": any_of_parse,
    "value": word_parse,
    "regex": regex_parse,
    "entity": entity_parse,
    "lemma": partial(not_supported, "LEMMA"),
    "pos": partial(not_supported, "POS"),
    "punct": punct_parse,
    "fuzzy": fuzzy_parse,
    "phrase": phrase_parse,
}


def rules_to_patterns(label, data):
    logger.debug("data: {}".format(data))

    def gen():
        """
        Implicitly add spaces between rules
        """
        if len(data) == 0:
            return

        yield data[0]

        for (t, d, op) in data[1:]:
            yield t, d, op

    return (
        label,
        [PARSERS[t](d, op) for (t, d, op) in gen()],
    )


class RuleExecutor(object):
    def __init__(self, patterns, config):
        self.config = config
        self.patterns = [self.compile(label, rules)
                         for label, rules in patterns]
        self.raw_patterns = patterns

    def compile(self, label, rules):
        flags = re.DOTALL
        if self.config.ignore_case:
            flags = flags | re.IGNORECASE

        regex_str = r"(?P<{0}>{1})".format(label, "".join(rules))
        try:
            return re.compile(regex_str, flags)
        except Exception as ex:
            logger.exception("Failed to compile: '{0}', Reason: \n{1}".format(regex_str, str(ex)))
            return None

    def _results(self, text):
        for p in self.patterns:
            for match in p.finditer(text):
                yield {
                    "start": match.start(),
                    "end": match.end(),
                    "text": match.group().strip(),
                    "label": match.lastgroup,
                }

    def execute(self, text):
        results = sorted(list(self._results(text)), key=lambda x: x["start"])
        for k, g in groupby(results, lambda x: x["start"]):
            group = list(g)
            if len(group) == 1:
                yield group[0]
            else:
                data = sorted(group, key=lambda x: -x["end"])
                yield data[0]

    @staticmethod
    def load(path):
        from rita.config import SessionConfig
        config = SessionConfig()
        with open(path, "r") as f:
            patterns = [(obj["label"], obj["rules"])
                        for obj in map(json.loads, f.readlines())]
            return RuleExecutor(patterns, config)

    def save(self, path):
        with open(path, "w") as f:
            for pattern in self:
                f.write("{0}\n".format(json.dumps(pattern)))

    def __iter__(self):
        for label, rules in self.raw_patterns:
            yield {"label": label, "rules": rules}


def compile_rules(rules, config):
    logger.info("Using standalone rule implementation")
    patterns = [rules_to_patterns(*group) for group in rules]
    executor = RuleExecutor(patterns, config)
    return executor



import re
import string

from rita.macros import resolve_value


char_translation = dict(
    [(c * 2, "{0}{{1,2}}".format(c)) for c in string.ascii_lowercase]
)

find_re = "|".join(["({0})".format(s) for (s, _) in char_translation.items()])

slang = {"you": "u", "for": "4", "are": "r", "you are": "ur", "you're": "ur"}


def premutations(initial):
    # return initial value
    yield initial

    """
    if we have double letters, like `oo`, we can guess that
    - user can sometimes enter both
    - sometimes only single
    """
    double_letters = re.sub(
        find_re,
        lambda x: char_translation[x.group(0)],
        initial
    )
    yield double_letters

    # if we have simple word, can add slang alternative
    if initial in slang:
        yield r"\s{0}\s".format(slang[initial])


def FUZZY(name, config, op=None):
    initial = resolve_value(name, config=config)
    return ("fuzzy", list(premutations(initial.lower())), op)
